"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.categoryIso = exports.invariantIso = exports.URI = exports.imap = exports.modifyF = exports.modify = exports.reverse = exports.compose = exports.asTraversal = exports.asOptional = exports.asPrism = exports.asLens = exports.id = void 0;
var function_1 = require("fp-ts/lib/function");
var O = require("fp-ts/lib/Option");
var _ = require("./internal");
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 2.3.0
 */
var id = function () { return ({
    get: function_1.identity,
    reverseGet: function_1.identity
}); };
exports.id = id;
// -------------------------------------------------------------------------------------
// converters
// -------------------------------------------------------------------------------------
/**
 * View an `Iso` as a `Lens`
 *
 * @category converters
 * @since 2.3.0
 */
exports.asLens = _.isoAsLens;
/**
 * View an `Iso` as a `Prism`
 *
 * @category converters
 * @since 2.3.0
 */
var asPrism = function (sa) { return ({
    getOption: function_1.flow(sa.get, O.some),
    reverseGet: sa.reverseGet
}); };
exports.asPrism = asPrism;
/**
 * View an `Iso` as a `Optional`
 *
 * @category converters
 * @since 2.3.0
 */
exports.asOptional = _.isoAsOptional;
/**
 * View an `Iso` as a `Traversal`
 *
 * @category converters
 * @since 2.3.0
 */
var asTraversal = function (sa) { return ({
    modifyF: function (F) { return function (f) { return function (s) { return F.map(f(sa.get(s)), function (a) { return sa.reverseGet(a); }); }; }; }
}); };
exports.asTraversal = asTraversal;
// -------------------------------------------------------------------------------------
// compositions
// -------------------------------------------------------------------------------------
/**
 * Compose an `Iso` with an `Iso`
 *
 * @category compositions
 * @since 2.3.0
 */
var compose = function (ab) { return function (sa) { return ({
    get: function_1.flow(sa.get, ab.get),
    reverseGet: function_1.flow(ab.reverseGet, sa.reverseGet)
}); }; };
exports.compose = compose;
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 2.3.0
 */
var reverse = function (sa) { return ({
    get: sa.reverseGet,
    reverseGet: sa.get
}); };
exports.reverse = reverse;
/**
 * @category combinators
 * @since 2.3.0
 */
var modify = function (f) { return function (sa) { return function (s) { return sa.reverseGet(f(sa.get(s))); }; }; };
exports.modify = modify;
function modifyF(F) {
    return function (f) { return function (sa) { return function (s) { return function_1.pipe(sa.get(s), f, function (fa) { return F.map(fa, sa.reverseGet); }); }; }; };
}
exports.modifyF = modifyF;
// -------------------------------------------------------------------------------------
// pipeables
// -------------------------------------------------------------------------------------
/**
 * @category Invariant
 * @since 2.3.0
 */
var imap = function (f, g) { return function (ea) {
    return imap_(ea, f, g);
}; };
exports.imap = imap;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
var imap_ = function (ea, ab, ba) { return ({
    get: function_1.flow(ea.get, ab),
    reverseGet: function_1.flow(ba, ea.reverseGet)
}); };
/**
 * @category instances
 * @since 2.3.0
 */
exports.URI = 'monocle-ts/Iso';
/**
 * @category instances
 * @since 2.3.0
 */
exports.invariantIso = {
    URI: exports.URI,
    imap: imap_
};
/**
 * @category instances
 * @since 2.3.0
 */
exports.categoryIso = {
    URI: exports.URI,
    compose: function (ab, ea) { return exports.compose(ab)(ea); },
    id: exports.id
};
