"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.categoryOptional = exports.invariantOptional = exports.URI = exports.imap = exports.findFirst = exports.traverse = exports.left = exports.right = exports.some = exports.atKey = exports.key = exports.index = exports.component = exports.props = exports.prop = exports.filter = exports.fromNullable = exports.modifyF = exports.modify = exports.modifyOption = exports.compose = exports.asTraversal = exports.id = void 0;
var function_1 = require("fp-ts/lib/function");
var O = require("fp-ts/lib/Option");
var pipeable_1 = require("fp-ts/lib/pipeable");
var _ = require("./internal");
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 2.3.0
 */
var id = function () { return ({
    getOption: O.some,
    set: function_1.constant
}); };
exports.id = id;
// -------------------------------------------------------------------------------------
// converters
// -------------------------------------------------------------------------------------
/**
 * View a `Optional` as a `Traversal`
 *
 * @category converters
 * @since 2.3.0
 */
exports.asTraversal = _.optionalAsTraversal;
// -------------------------------------------------------------------------------------
// compositions
// -------------------------------------------------------------------------------------
/**
 * Compose a `Optional` with a `Optional`
 *
 * @category compositions
 * @since 2.3.0
 */
exports.compose = _.optionalComposeOptional;
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * @category combinators
 * @since 2.3.0
 */
exports.modifyOption = _.optionalModifyOption;
/**
 * @category combinators
 * @since 2.3.0
 */
exports.modify = _.optionalModify;
function modifyF(F) {
    return function (f) { return function (sa) { return function (s) {
        return pipeable_1.pipe(sa.getOption(s), O.fold(function () { return F.of(s); }, function (a) { return F.map(f(a), function (a) { return sa.set(a)(s); }); }));
    }; }; };
}
exports.modifyF = modifyF;
/**
 * Return an `Optional` from a `Optional` focused on a nullable value
 *
 * @category combinators
 * @since 2.3.3
 */
exports.fromNullable = 
/*#__PURE__*/
exports.compose(_.prismAsOptional(_.prismFromNullable()));
function filter(predicate) {
    return exports.compose(_.prismAsOptional(_.prismFromPredicate(predicate)));
}
exports.filter = filter;
/**
 * Return a `Optional` from a `Optional` and a prop
 *
 * @category combinators
 * @since 2.3.0
 */
var prop = function (prop) {
    return exports.compose(pipeable_1.pipe(_.lensId(), _.lensProp(prop), _.lensAsOptional));
};
exports.prop = prop;
/**
 * Return a `Optional` from a `Optional` and a list of props
 *
 * @category combinators
 * @since 2.3.0
 */
var props = function () {
    var props = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        props[_i] = arguments[_i];
    }
    return exports.compose(pipeable_1.pipe(_.lensId(), _.lensProps.apply(_, props), _.lensAsOptional));
};
exports.props = props;
/**
 * Return a `Optional` from a `Optional` and a component
 *
 * @category combinators
 * @since 2.3.0
 */
var component = function (prop) {
    return exports.compose(pipeable_1.pipe(_.lensId(), _.lensComponent(prop), _.lensAsOptional));
};
exports.component = component;
/**
 * Return a `Optional` from a `Optional` focused on a `ReadonlyArray`
 *
 * @category combinators
 * @since 2.3.0
 */
var index = function (i) { return function (sa) {
    return pipeable_1.pipe(sa, _.optionalComposeOptional(_.indexArray().index(i)));
}; };
exports.index = index;
/**
 * Return a `Optional` from a `Optional` focused on a `ReadonlyRecord` and a key
 *
 * @category combinators
 * @since 2.3.0
 */
var key = function (key) { return function (sa) {
    return pipeable_1.pipe(sa, _.optionalComposeOptional(_.indexRecord().index(key)));
}; };
exports.key = key;
/**
 * Return a `Optional` from a `Optional` focused on a `ReadonlyRecord` and a required key
 *
 * @category combinators
 * @since 2.3.0
 */
var atKey = function (key) { return function (sa) {
    return pipeable_1.pipe(sa, exports.compose(_.lensAsOptional(_.atRecord().at(key))));
}; };
exports.atKey = atKey;
/**
 * Return a `Optional` from a `Optional` focused on the `Some` of a `Option` type
 *
 * @category combinators
 * @since 2.3.0
 */
exports.some = 
/*#__PURE__*/
exports.compose(_.prismAsOptional(_.prismSome()));
/**
 * Return a `Optional` from a `Optional` focused on the `Right` of a `Either` type
 *
 * @category combinators
 * @since 2.3.0
 */
exports.right = 
/*#__PURE__*/
exports.compose(_.prismAsOptional(_.prismRight()));
/**
 * Return a `Optional` from a `Optional` focused on the `Left` of a `Either` type
 *
 * @category combinators
 * @since 2.3.0
 */
exports.left = 
/*#__PURE__*/
exports.compose(_.prismAsOptional(_.prismLeft()));
/**
 * Return a `Traversal` from a `Optional` focused on a `Traversable`
 *
 * @category combinators
 * @since 2.3.0
 */
function traverse(T) {
    return function_1.flow(exports.asTraversal, _.traversalComposeTraversal(_.fromTraversable(T)()));
}
exports.traverse = traverse;
function findFirst(predicate) {
    return exports.compose(_.findFirst(predicate));
}
exports.findFirst = findFirst;
// -------------------------------------------------------------------------------------
// pipeables
// -------------------------------------------------------------------------------------
/**
 * @category Invariant
 * @since 2.3.0
 */
var imap = function (f, g) { return function (ea) { return imap_(ea, f, g); }; };
exports.imap = imap;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
var imap_ = function (ea, ab, ba) { return ({
    getOption: function_1.flow(ea.getOption, O.map(ab)),
    set: function_1.flow(ba, ea.set)
}); };
/**
 * @category instances
 * @since 2.3.0
 */
exports.URI = 'monocle-ts/Optional';
/**
 * @category instances
 * @since 2.3.0
 */
exports.invariantOptional = {
    URI: exports.URI,
    imap: imap_
};
/**
 * @category instances
 * @since 2.3.0
 */
exports.categoryOptional = {
    URI: exports.URI,
    compose: function (ab, ea) { return exports.compose(ab)(ea); },
    id: exports.id
};
